import os.path
import re
from typing import Dict, Optional, Union, Iterable, Tuple, List, IO, Pattern
from dataclasses import dataclass
import lxml.etree as ET

SegmontoZones = frozenset(["CustomZone",
                           "DamageZone",
                           "GraphicZone",
                           "DigitizationArtefactZone",
                           "DropCapitalZone",
                           "MainZone",
                           "MarginTextZone",
                           "MusicZone",
                           "NumberingZone",
                           "QuireMarksZone",
                           "RunningTitleZone",
                           "SealZone",
                           "StampZone",
                           "TableZone",
                           "TitlePageZone"])

SegmontoLines = frozenset(["CustomLine",
                           "DefaultLine",
                           "DropCapitalLine",
                           "HeadingLine",
                           "InterlinearLine",
                           "MusicLine"])

SegmontoZoneRegex: Pattern = re.compile(f"({'|'.join(SegmontoZones)})" + r"(:\w+)?(#\w+)?")
SegmontoLineRegex: Pattern = re.compile(f"({'|'.join(SegmontoLines)})" + r"(:\w+)?(#\w+)?")


@dataclass
class Element:
    id: str
    tagname: str
    category: Optional[str] = None
    has_content: bool = False


class XmlParser:
    def parse(self):
        raise NotImplemented

    def get_zones(self) -> Iterable[Element]:
        raise NotImplemented

    def get_textlines(self) -> Iterable[Element]:
        raise NotImplemented

    def test(
            self,
            check_empty: bool = False,
            check_typing: bool = False,
            typing_check_zones: Optional[Pattern] = None,
            typing_check_lines: Optional[Pattern] = None
    ) -> Tuple[List[Element], List[Element], List[Element]]:
        zones_error = []
        line_error = []
        empty = []
        for zone in self.get_zones(check_empty=check_empty):
            if check_typing and typing_check_zones:
                if (zone.category and not typing_check_zones.match(zone.category)) or not zone.category:
                    zones_error.append(zone)
            if not zone.has_content and check_empty:
                empty.append(zone)
        for line in self.get_textlines(check_empty=check_empty):
            if check_typing and typing_check_lines:
                if (line.category is not None and not typing_check_lines.match(line.category)) or line.category is None:
                    line_error.append(line)
            if not line.has_content and check_empty:
                empty.append(line)
        return zones_error, line_error, empty

    def _check_line_content(self, line: ET._Element) -> bool:
        raise NotImplemented

    def _check_zone_content(self, zone: ET._Element) -> bool:
        raise NotImplemented

    def check_image_link(self, filepath: Optional[str] = None) -> Tuple[str, bool]:
        raise NotImplementedError

    def _check_image_link(self, filepath: Optional[str], xpath_results: Iterable[str]) -> Tuple[str, bool]:
        if not filepath:
            raise FileNotFoundError("Can't check an image link without a filepath")
        for filename in xpath_results:
            filename = str(filename)
            filename = os.path.join(os.path.dirname(filepath), filename)
            return filename, os.path.exists(filename)
        return "", False


class PageXML(XmlParser):
    def __init__(self, file: Union[str, ET._ElementTree, IO]):
        if isinstance(file, str):
            self.xml = ET.parse(file)
        else:
            self.xml = file

    @staticmethod
    def _parse_custom(attribute_string):
        annotations = {}
        attribute_string = attribute_string.strip()
        annotations_chunks = [l_chunk for l_chunk in attribute_string.split('}') if l_chunk.strip() and "{" in l_chunk]
        if annotations_chunks:
            for chunk in annotations_chunks:
                tag, vals = chunk.split('{')
                tag_vals = {}
                vals = [val.strip() for val in vals.split(';') if val.strip()]
                for val in vals:
                    key, *val = val.split(':')
                    tag_vals[key] = ":".join(val)
                annotations[tag.strip()] = tag_vals
        return annotations.get("structure", {}).get("type", None)

    def get_zones(self, check_empty: bool = False):
        for region in self.xml.findall(".//{*}TextRegion"):
            yield Element(
                id=region.attrib.get("id", "UnknownID"), tagname="Region",
                category=self._parse_custom(region.attrib.get("custom", "")),
                has_content=False if not check_empty else self._check_zone_content(region)
            )

    def _check_zone_content(self, zone: ET._Element) -> bool:
        return zone.find(".//{*}TextLine") is not None

    def get_textlines(self, check_empty: bool = False):
        for line in self.xml.findall('.//{*}TextLine'):
            yield Element(
                id=line.get("id", "UnknownID"), tagname="Line",
                category=self._parse_custom(line.attrib.get("custom", "")),
                has_content=False if not check_empty else self._check_line_content(line)
            )

    def _check_line_content(self, line: ET._Element) -> bool:
        _line = line.find(".//{*}Unicode")
        if _line is not None:
            if _line.text is None:
                return False
            return bool(_line.text.strip())
        return False

    def check_image_link(self, filepath: Optional[str] = None) -> Tuple[str, bool]:
        return self._check_image_link(filepath, self.xml.xpath("//@imageFilename"))


class AltoXML(XmlParser):
    _Regions = {'TextBlock': 'text',
                'IllustrationType': 'illustration',
                'GraphicalElementType': 'graphic',
                'ComposedBlock': 'composed'}

    def __init__(self, file: Union[str, ET._ElementTree, IO]):
        if isinstance(file, str):
            self.xml = ET.parse(file)
        else:
            self.xml = file
        self._classes = self._get_class_maps(self.xml)

    def _parse_tagrefs(self, attribute_string: str) -> Optional[str]:
        for tagref in attribute_string.split():
            rtype = self._classes.get(tagref, None)
            if rtype:
                return rtype
        return

    def _get_class_maps(self, doc) -> Dict[str, str]:
        cls_map = {}
        tags = doc.find('.//{*}Tags')
        if tags is not None:
            for x in ['StructureTag', 'LayoutTag', 'OtherTag']:
                for tag in tags.findall('./{{*}}{}'.format(x)):
                    cls_map[tag.get('ID')] = tag.get('LABEL')
        return cls_map

    def get_textlines(self, check_empty: bool = False):
        for line in self.xml.findall('.//{*}TextLine'):
            yield Element(
                id=line.get("ID", "UnknownID"), tagname="Line",
                category=self._parse_tagrefs(line.get('TAGREFS', "")),
                has_content=False if not check_empty else self._check_line_content(line)
            )

    def get_zones(self, check_empty: bool = False):
        regions = []
        for x in AltoXML._Regions.keys():
            regions.extend(self.xml.findall('./{{*}}Layout/{{*}}Page/{{*}}PrintSpace/{{*}}{}'.format(x)))
        for region in regions:
            yield Element(
                id=region.get("ID", "UnknownID"), tagname="Region",
                category=self._parse_tagrefs(region.get('TAGREFS', "")),
                has_content=False if not check_empty else self._check_zone_content(region)
            )

    def _check_line_content(self, line: ET._Element) -> bool:
        _line = line.find("{*}String")
        if _line is not None:
            return bool(_line.attrib["CONTENT"].strip())
        return False

    def _check_zone_content(self, zone: ET._Element) -> bool:
        return zone.find(".//{*}TextLine") is not None

    def check_image_link(self, filepath: Optional[str] = None) -> Tuple[str, bool]:
        return self._check_image_link(
            filepath,
            [el.text for el in self.xml.findall("//{*}fileName")]
        )



import os
import requests
import hashlib
from typing import Optional, Union
from lxml import etree
import logging

logger = logging.getLogger(__name__)

_here = os.path.dirname(__file__)


Schemas = {
    "alto": os.path.join(_here, "", "alto4.xsd"),
    "page": os.path.join(_here, "", "page2019.xsd")
}


class CacheResolver(etree.Resolver):
    cache = _here

    def resolve(self, URL, id, context):
        # determine cache path
        if os.path.exists(URL):
            return self.resolve_file(open(URL), context, base_url=URL)
        xsd_path = Validator.cache_xsd_path(URL)
        if os.path.exists(xsd_path):
            return self.resolve_file(open(xsd_path), context, base_url=URL)
        raise requests.HTTPError(f"Unable to reach {URL} and not found in cache at {xsd_path}")


class Validator:
    def __init__(self, xsd_path: str):
        parser = etree.XMLParser(no_network=True)
        parser.resolvers.add(CacheResolver())
        xmlschema_doc = etree.parse(self.get_schema(xsd_path), parser=parser)
        self.xmlschema = etree.XMLSchema(xmlschema_doc)

    @staticmethod
    def retrieve_xsd(file: Union[str, etree._ElementTree]) -> Optional[str]:
        ns = '{http://www.w3.org/2001/XMLSchema-instance}'
        if not isinstance(file, etree._ElementTree):
            document = etree.parse(file)
        else:
            document = file
        schemaLink = document.getroot().get(ns + 'schemaLocation')
        if schemaLink:
            for link in schemaLink.split():
                if link.endswith(".xsd"):
                    return Validator.get_schema(link)
        return None

    @staticmethod
    def cache_xsd_path(xsd_path):
        xsd_path = xsd_path.replace("http://", "").replace("https://", "")
        return os.path.join(_here, "", f"{hashlib.sha256(xsd_path.encode()).hexdigest()}.xsd")

    @staticmethod
    def get_schema(xsd_path):
        if xsd_path.startswith("http://") or xsd_path.startswith("https://"):
            new_xsd_path = f"downloaded_{hashlib.sha256(xsd_path.encode()).hexdigest()}.xsd"
            if os.path.exists(Validator.cache_xsd_path(xsd_path)):
                return Validator.cache_xsd_path(xsd_path)
            elif os.path.exists(new_xsd_path):
                return new_xsd_path
            try:
                schema_req = requests.get(xsd_path)
                schema_req.raise_for_status()
            except requests.HTTPError as E:
                logger.warning(f"HTTP error while contacting {xsd_path}: {E}. Trying HTTPS if HTTP failed.")
                # ALTO seems to throw an error because they moved to HTTP
                if xsd_path.startswith("http://"):
                    return Validator.get_schema(xsd_path.replace("http://", "https://"))
                raise
            with open(new_xsd_path, "w") as f:
                f.write(schema_req.text)
            return new_xsd_path

        elif xsd_path in Schemas:
            return Schemas[xsd_path]
        return xsd_path

    def validate(self, xml_path: Union[str, etree._ElementTree]) -> bool:
        if not isinstance(xml_path, etree._ElementTree):
            xml_doc = etree.parse(xml_path)
        else:
            xml_doc = xml_path
        result = self.xmlschema.validate(xml_doc)
        return result


def simplify_log_line(string: etree._LogEntry) -> str:
    return string.message.replace("{http://www.loc.gov/standards/alto/ns-v4#}", "alto:")\
        .replace("{http://schema.primaresearch.org/PAGE/gts/pagecontent/2019-07-15}", "page:")





import re
from collections import defaultdict
from typing import Iterable, List, Dict, Optional, Tuple, Union, IO, Sequence
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import click
from lxml import etree

from htrvx.schemas import Validator, simplify_log_line
from htrvx.zones import AltoXML, PageXML, Element, SegmontoZoneRegex, SegmontoLineRegex
from dataclasses import dataclass

# Spacing for printing
Space1 = "  "
Space2 = "    "


def _color(status, mode: Optional[str] = None):
    if mode == "zen":
        if status == "failure":
            return "red"
        else:
            return None

    if status == "success":
        return "green"
    elif status == "warning":
        return "yellow"
    else:
        return "red"


def _char(status):
    if status == "success":
        return "✓"
    elif status == "warning":
        return "⚠"
    else:
        return "×"


def _msg(status):
    if status == "success":
        return "passed successfully"
    elif status == "warning":
        return "has warnings"
    else:
        return "failed"


@dataclass
class Status:
    status: Literal["success", "warning", "failure"]
    task: Literal["segmonto", "schema", "empty-verification", "image-link-check", "custom-typing-check"]
    message: Optional[str] = None
    errors: Optional[List[str]] = None
    level: Optional[Literal["zone", "line"]] = None

    def print(self, mode: Optional[str] = None) -> None:
        if mode in {"minimal", "low"} and self.status == "success":
            return None
        additional_info = ""
        if self.level:
            additional_info = f" at the \033[1m{self.level}\033[0m's level"

        click.echo(
            click.style(
                f"{Space1}{_char(self.status)} \033[1m{' '.join(self.task.capitalize().split('-'))}\033[0m's "
                f"test{additional_info} {_msg(status=self.status)}{': '+self.message if self.message else ''}.",
                fg=_color(self.status, mode=mode)
            ),
            color=True
        )
        if self.errors and mode != "minimal":
            for error in self.errors:
                click.echo(click.style(f"{Space2}┗ {error}", fg="blue" if mode != "zen" else None), color=True)


@dataclass
class FileLog:
    tests: Optional[List[Status]] = None

    def append(self, value) -> None:
        if self.tests is None:
            self.tests = []
        self.tests.append(value)

    @property
    def status(self) -> bool:
        passed, total = self.score
        return passed == total

    @property
    def score(self) -> Tuple[int, int]:
        """Returns the number of passing test and test done"""
        statuses = [
            0 if file_status.status == "failure" else 1
            for file_status in self.tests
        ] if self.tests else []
        return sum(statuses), len(self.tests or [])

    def __iter__(self) -> Iterable[Status]:
        return iter(self.tests) if self.tests else []

    def __len__(self) -> int:
        return len(self.tests) if self.tests else 0

    def __bool__(self) -> bool:
        return self.status

    def print(self, mode: Optional[str] = None) -> None:
        if self.tests:
            for element in self.tests:
                element.print(mode=mode)


def _empty_or_wrong(category):
    if category is None:
        return "is not categorized"
    return f"has a forbidden type (`{category}`)"


def _get_ids(ids: Iterable[str]) -> str:
    return ", ".join([f"#{val}" for val in ids])


def parse_segmonto_errors(errors: Iterable[Element], group=False, element_type="element") -> List[str]:
    """ Parse a list of errors
    
    """
    if not errors:
        return []

    if not group:
        return [
            f"{element_type.capitalize()} with id #{error.id} {_empty_or_wrong(error.category)}"
            for error in errors
        ]

    groups = defaultdict(list)
    for error in errors:
        groups[f"`{error.category}`" if error.category is not None else "Missing"].append(error.id)

    return [
        f"{tag_name} tag for {element_type}(s) is forbidden ({len(ids)} annotations): {_get_ids(ids)}"
        for tag_name, ids in sorted(list(groups.items()), key=lambda x: x[0])
    ]


def parse_empty(empty, group=False) -> Tuple[List[str], List[str]]:
    """ Parses the empty log and returns two list of errors: one for lines, one for regions

    """
    if not empty:
        return [], []

    zone_errors = [z.id for z in empty if z.tagname == "Region"]
    line_errors = [z.id for z in empty if z.tagname == "Line"]

    if not group:
        return (
            [f"Region with id #{zone} is empty" for zone in zone_errors],
            [f"Line with id #{line} is empty" for line in line_errors]
        )
    return (
        [f"Zones with missing IDs: {_get_ids(zone_errors)}"] if zone_errors else [],
        [f"Lines with missing IDs: {_get_ids(line_errors)}"] if line_errors else []
    )


def parse_alto_logs(error_log: Iterable[etree._LogEntry], group: bool = False) -> List[str]:
    """ Parses a Schema error log and returns the error with simplifications as string

    """
    if not group:
        return [
            f"Line {line.line:04d}: {simplify_log_line(line)}"
            for line in error_log
        ]
    errors = defaultdict(list)
    for line in error_log:
        if group:
            errors[simplify_log_line(line)].append(str(line.line))

    return [
        f"{error} on line(s): {', '.join(lines)}"
        for error, lines in errors.items()
    ]


def test_single(
    file: Union[str, IO, etree._ElementTree],
    group: bool = True,
    format: Literal["alto", "page"] = "alto",
    segmonto: bool = True,
    check_empty: bool = True,
    raise_empty: bool = True,
    xsd: bool = False,
    check_image: bool = False,
    zones: Optional[Sequence[str]] = None,
    lines: Optional[Sequence[str]] = None
) -> FileLog:
    filelog = FileLog()

    if format == "alto":
        cls = AltoXML
    elif format == "page":
        cls = PageXML
    else:
        raise ValueError("Format for files should be either `alto` or `page`")

    if not hasattr(file, "xpath"):  # Definitely not perfect, ToDo: FIX
        parsed_xml = etree.parse(file)
    else:
        parsed_xml = file

    custom_typing_check = bool(zones or lines)

    line_regex = None
    zone_regex = None
    if custom_typing_check:
        if lines:
            line_regex = re.compile("|".join([re.escape(pat) for pat in lines]))
        if zones:
            zone_regex = re.compile("|".join([re.escape(pat) for pat in zones]))
    elif segmonto:
        line_regex = SegmontoLineRegex
        zone_regex = SegmontoZoneRegex

    # For some tests, we need to parse the file internally
    if segmonto or check_empty or check_image or custom_typing_check:
        obj = cls(parsed_xml)

        if check_image:
            filepath, status = obj.check_image_link(file if isinstance(file, str) else None)
            message = ""
            if not status:
                if filepath:
                    message = f"Image file at path `{filepath}` not found."
                else:
                    message = "No image file were declared in the XML."

            filelog.append(Status(
                "success" if status else "failure",
                task="image-link-check",
                message=message if message else None
            ))

        zone_errors, line_errors, empty = obj.test(
            check_empty=check_empty,
            check_typing=segmonto or custom_typing_check,
            typing_check_lines=line_regex,
            typing_check_zones=zone_regex
        )

        if segmonto or custom_typing_check:
            if segmonto or zones:
                filelog.append(
                    Status(
                        "success" if not zone_errors else "failure",
                        task="segmonto" if segmonto else "custom-typing-check",
                        message=f"{len(zone_errors)} wrongly tagged zones" if zone_errors else "",
                        errors=parse_segmonto_errors(zone_errors, group=group, element_type="zone"),
                        level="zone"
                    )
                )
            if segmonto or lines:
                filelog.append(
                    Status(
                        "success" if not line_errors else "failure",
                        task="segmonto" if segmonto else "custom-typing-check",
                        message=f"{len(line_errors)} wrongly tagged lines" if line_errors else "",
                        errors=parse_segmonto_errors(line_errors, group=group, element_type="line"),
                        level="line"
                    )
                )

        if check_empty:
            empty = parse_empty(empty, group=group)

            for results, element_type in zip(empty, ["zone", "line"]):
                if not results:
                    success = "success"
                elif raise_empty and results:
                    success = "failure"
                else:
                    success = "warning"
                filelog.append(
                    Status(
                        success,
                        task="empty-verification",
                        message=f"{len(results)} empty {element_type}(s) found" if results else "",
                        errors=results,
                        level=element_type
                    )
                )
    if xsd:
        validator = Validator(Validator.retrieve_xsd(parsed_xml))
        if validator.validate(parsed_xml):
            filelog.append(Status("success", task="schema", message="validation passed"))
        else:
            filelog.append(
                Status(
                    "failure",
                    task="schema",
                    message="validation failed",
                    errors=parse_alto_logs(validator.xmlschema.error_log, group=group)
                )
            )

    return filelog


def test(
    files: Iterable[Union[str, IO, etree._ElementTree]],
    verbose: bool = False,
    group: bool = True,
    format: str = "alto",
    segmonto: bool = True,
    check_empty: bool = True,
    raise_empty: bool = False,
    check_image: bool = False,
    xsd: bool = False,
    verbose_level: str = "all",
    zones: Optional[Sequence[str]] = None,
    lines: Optional[Sequence[str]] = None
) -> Tuple[Dict[str, FileLog], bool]:
    """ Tests all single files in files and returns their filelog as well as a global boolean status

    """
    statuses: Dict[str, FileLog] = defaultdict(FileLog)

    for idx, file in enumerate(files):
        if not isinstance(file, str):
            file_name = "File %s" % str(idx+1).zfill(3)
        else:
            file_name = file

        statuses[file_name] = test_single(
            file,
            group=group, format=format,
            segmonto=segmonto, check_empty=check_empty, raise_empty=raise_empty, xsd=xsd,
            check_image=check_image, zones=zones, lines=lines
        )
        if verbose:
            filelog = statuses[file_name]
            passed, total = filelog.score
            status_string: str = "success" if passed == total else "failure"
            # Print the element overal status
            click.echo(
                click.style(
                    f"{_char(status_string)} [{passed}/{total}] {file_name}",
                    fg=_color(status_string, mode=verbose_level)
                ),
                color=True
            )
            if status_string == "success" and verbose_level in {"minimal", "low"}:
                continue
            # Print the details
            filelog.print(mode=verbose_level)

    passing_files = [int(bool(file_statuses)) for file_statuses in statuses.values()]

    if verbose:
        click.echo("\n\n\n=====\nREPORT\n=====\n")
        click.echo(f"{sum(passing_files)}/{len(statuses)} valid XML files")

    return statuses, len(statuses) == sum(passing_files)




import sys
import click

from htrvx.testing import test
from typing import Sequence, Optional

@click.command()
@click.argument("files", nargs=-1, type=click.Path(exists=True, dir_okay=False, file_okay=True))
@click.option("-v", "--verbose", default=False, is_flag=True,
              help="Prints more information", show_default=True)
@click.option("-f", "--format", default="alto", type=click.Choice(["alto", "page", "auto"]),
              help="Format of files", show_default=True)
@click.option("-s", "--segmonto", is_flag=True, default=False,
              help="Apply Segmonto Zoning verification", show_default=True)
@click.option("--zone", default=None, multiple=True,
              help="Provide a custom zone to control zone types instead of Segmonto", show_default=True)
@click.option("--line", default=None, multiple=True,
              help="Provide a custom zone to control lines types instead of Segmonto", show_default=True)
@click.option("-e", "--check-empty", is_flag=True, default=False,
              help="Check for empty lines or empty zones", show_default=True)
@click.option("-i", "--check-image", is_flag=True, default=False,
              help="Check if image links in the XML points to real files", show_default=True)
@click.option("-r", "--raise-empty", is_flag=True, default=False,
              help="Warns but not fails if empty lines or empty zones are found", show_default=True)
@click.option("-x", "--xsd", is_flag=True, default=False,
              help="Apply XSD Schema verification", show_default=True)
@click.option("-l", "--verbose-level", default="zen", type=click.Choice(["minimal", "low", "zen", "all"]),
              help="Verbosity level. Minimal show only failing test, "
                   "zen only displays one color (red), "
                   "low shows failing tests with their debug info, "
                   "all shows everything", show_default=True)
@click.option("-g", "--group", default=False, is_flag=True, help="Group error types")
def cmd(files, verbose: bool = False, group: bool = True, format: str ="alto", segmonto: bool = True,
        check_empty: bool = True, raise_empty: bool = True,
        xsd: bool = False, check_image: bool = False, verbose_level: str = "zen",
        zone: Optional[Sequence[str]] = None, line: Optional[Sequence[str]] = None):
    """ Apply the XSD on FILES. XSD can be a URI, a filepath or a schema provided with this tool (eg. "ALTO-Segmonto")

    eg. `htrvx ./data/**/*.xml --group --schema --format alto`

    With multiple zones

    eg. `htrvx ./data/**/*.xml --group --schema --format alto --zone Col --zone Header`

    """
    if test(files, verbose=verbose, group=group, format=format, segmonto=segmonto,
            xsd=xsd, raise_empty=raise_empty, check_empty=check_empty, check_image=check_image,
            verbose_level=verbose_level, zones=zone, lines=line)[1]:
        sys.exit(0)
    else:
        sys.exit(1)


if __name__ == "__main__":
    cmd()
